/*

    File: iterative parallax.frag

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
varying vec3 lightVec;
varying vec3 eyeVec;
varying vec2 texCoord;
uniform sampler2D colorMap;
uniform sampler2D normalMap;
uniform float polyThickness;

// ******************************** flag defs ********************************

//enable this to enable parallax mapping at all
//all other preprocessors depend on it
//#define PARALLAX

//depends on PARALLAX
//use normalized x & y for computing the offset vector
//that way we don't extend further than our height component
//and thus reciprocals don't extend out forever
//When you use this with the ITERATIVE effect, it gives an incorrect fish-eye look
//#define OFFSET_LIMITING_EYE

//depends on PARALLAX
//try to step through the shader
//this is difficult considering how our cards can only handle 4 texture lookups.
//but works well in Education - 32 isn't even the upper limit!
//#define ITERATIVE

//number of steps through a medium
//this value is supposed to be dynamic.
//maybe later. . . 
#ifndef ITERATIVE_STEPS
#define ITERATIVE_STEPS			32.0
#endif //ITERATIVE_STEPS

//step from 0.0 to 1.0
//1.0 is the most you can get out of the Graphics lab cards
//of course, those don't even support GLSL shader loops,
//so dont worry about them in the ITERATIVE option:
#define ITERATIVE_STEP_SIZE		(1.0 / ITERATIVE_STEPS)

//depends on ITERATIVE
//enable this to add linear interpolation to the collision point
//removes jagged artifacts
//#define ITERATIVE_LERP

//depends on ITERATIVE
//traces back from the collision point towards the light source
// ... or from the light source to the collision point?
//we'll see =)
//#define SHADOW_TRACE

//Whether to use offset limiting on calculating the shadow slope as well
//honestly for these iterative methods you're
//(a) not supposed to use offset limiting at all and
//(b) supposed to vary the step size by the distance the ray covers. whoops =)
//#define OFFSET_LIMITING_LIGHT

// ******************************** the code ********************************

void main (void) {
	
	float lightDistSqr = dot(lightVec,lightVec);
	float lightDist = sqrt(lightDistSqr);
	
	//3% decrease for using division instead of multiplication
	float attenuation = 1.0 / (gl_LightSource[0].constantAttenuation +
								gl_LightSource[0].linearAttenuation * lightDist +
								gl_LightSource[0].quadraticAttenuation * lightDistSqr);

	vec3 unitLightVec = lightVec / lightDist;

	vec3 unitEyeVec = normalize(eyeVec);
	
	//jumping into the shader option fray:
	//all options need to return two variables:
	//"bumpTexel" and "colorMapTexel"
	
#ifndef PARALLAX

	vec4 bumpTexel = texture2D(normalMap, texCoord);
	vec4 colorMapTexel = texture2D(colorMap, texCoord);						//map directly through

#else //PARALLAX

#ifdef OFFSET_LIMITING_EYE	//with offset limiting we remove the z divisor, thus giving us a circle where we would have a triangle
#define EYE_SLOPE_VECTOR	polyThickness * unitEyeVec.xy
#else //OFFSET_LIMITING_EYE
#define EYE_SLOPE_VECTOR	polyThickness / unitEyeVec.z * unitEyeVec.xy
#endif //OFFSET_LIMITING_EYE

	
#ifndef ITERATIVE

	vec4 bumpTexel = texture2D(normalMap, texCoord);

	//bumpmap height value of 1.0 corresponds to being at polygon value
	//this way we can 1-1 with our iterative method, which would get some clipping effects if we went over of the polygon surface
	vec2 offsetTexCoord = texCoord + (bumpTexel.w - 1.0) * EYE_SLOPE_VECTOR;

	//now re-grab out bumpmap texel from the offset coordinate
	bumpTexel = texture2D(normalMap, offsetTexCoord);

#else	//ITERATIVE
	//this option is only supported at the Education lab's cards:
	
	//the "unitEyeVec.z" signifies offset limiting -- should it be there for the iterative techinque
	vec2 deltaTexCoord = ITERATIVE_STEP_SIZE * EYE_SLOPE_VECTOR;

	//start at the start and trace forward
	vec2 offsetTexCoord = texCoord;

	//do our first height read:
	vec4 bumpTexel = texture2D(normalMap, offsetTexCoord);
	
	//if the first height read isn't on the top of the surface . . .
	if (1.0 > bumpTexel.w) {
		//then jump into it

		//for which know that our depth is gonna be at the top
		float currentDepth = 1.0;
	
		//and cycle through the loop with it:
		do {
			offsetTexCoord -= deltaTexCoord;
			currentDepth -= ITERATIVE_STEP_SIZE;
			bumpTexel = texture2D(normalMap, offsetTexCoord);
		} while (currentDepth > bumpTexel.w);

#ifdef SHADOW_TRACE
		//NOTICE: will the last LERP operation mess us up? should we step back before that?  I'm gonna
		vec2 lightTexCoord = offsetTexCoord + deltaTexCoord;
#endif //SHADOW_TRACE
		
#ifdef ITERATIVE_LERP
		//by now currentDepth has finally run beneath our bumpTexel . . .
		//if so, now we calculate our backstep fraction
		//we divide the dy/dx, and dx is ITERATIVE_STEP_SIZE, and 1/ITERATIVE_STEP_SIZE is ITERATIVE_STEPS . . .
		float backFrac = (bumpTexel.w - currentDepth) * ITERATIVE_STEPS;
		//offset our texcoord again
		offsetTexCoord += deltaTexCoord * backFrac;
		//and get the LERPd bump texel
		bumpTexel = texture2D(normalMap, offsetTexCoord);
#endif //ITERATIVE_LERP

#ifdef SHADOW_TRACE
		//bump currentDepth back a step so its above ground - and coinciding with the lightTexCoord
		currentDepth += ITERATIVE_STEP_SIZE;
		
#ifdef OFFSET_LIMITING_LIGHT
#define LIGHT_SLOPE_VECTOR polyThickness * unitLightVec.xy
#else //OFFSET_LIMITING_LIGHT
#define LIGHT_SLOPE_VECTOR polyThickness / unitLightVec.z * unitLightVec.xy
#endif //OFFSET_LIMITING_LIGHT
		
		//calc the delta tex coord from where we're at to the light source:
		vec2 lightDeltaTexCoord = ITERATIVE_STEP_SIZE * LIGHT_SLOPE_VECTOR;	//NOTICE  - SLOPE_VECTOR is influenced by OFFSET_LIMIT

		//dont grab over bumpTexel, so just grab the 'w' component of the bumpmap texel for light testing
		float lightBumpTexelW;
		
		//here's whether we made it to the light: assume worse at first:
		float selfShadowCoeff = 0.0;
		
		//muck up currentDepth -- no one's using it
		do {
			lightTexCoord += lightDeltaTexCoord;
			currentDepth += ITERATIVE_STEP_SIZE;
			if (currentDepth >= 1.0) {	//we made it to the light source!
				selfShadowCoeff = 1.0;	//set the flag!
				break;
			}
			lightBumpTexelW = texture2D(normalMap, lightTexCoord).w;
			
			selfShadowCoeff += ITERATIVE_STEP_SIZE;
			
		} while (currentDepth > lightBumpTexelW);
		
		//now tweak our attenuation to do the selfshadowing later
		attenuation *= selfShadowCoeff;
		
#endif //SHADOW_TRACE

	}
#endif	//ITERATIVE

	//by here all parallax based options have calculated offsetTexCoord
	//and all still need their colorMapTexel:
	vec4 colorMapTexel = texture2D(colorMap, offsetTexCoord);

#endif //PARALLAX

	vec3 bumpNormal = normalize( bumpTexel.xyz * 2.0 - 1.0);

	vec4 ambientColor = gl_LightSource[0].ambient * gl_FrontMaterial.ambient;

	float diffuseCoeff = max( dot(unitLightVec, bumpNormal), 0.0 );
	vec4 diffuseColor = gl_LightSource[0].diffuse * gl_FrontMaterial.diffuse * diffuseCoeff;	

	float specularCoeff = pow(clamp(dot(reflect(-unitEyeVec, bumpNormal), unitLightVec), 0.0, 1.0), gl_FrontMaterial.shininess);
	vec4 specularColor = gl_LightSource[0].specular * gl_FrontMaterial.specular * specularCoeff;	
	
	gl_FragColor = ambientColor + (diffuseColor * colorMapTexel + specularColor) * attenuation;
}