/*

    File: ShaderScene.cpp

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <assert.h>

#include <iostream>
#include <sstream>

#include "math/mat.h"

#include "system/singletons.h"	//kbd
#include "system/exception.h"
#include "base/main.h"
#include "base/filebuffer.h"

#include "menu/sys.h"
#include "menu/lib.h"

#include "lib/bumpmap.h"
#include "shader/shaderext.h"	//shaderSystem.init
#include "shader/program.h"

#include "gc/gc.h"

#ifdef WIN32
#define ALLOW_DRAG_AND_DROP
#endif

#ifdef ALLOW_DRAG_AND_DROP
#include "winsystem/singletons.h"
#include "winsystem/wstring.h"
#endif

using namespace std;
using namespace pwn;


//3D model, for our shader, needs special vertex data <TNB>, so im making a new class
class Model3D {
public:

	class Vertex {
	public:
		vec3f pos;
		vec2f tc;
		mat33f basis;	//tangent, normal, binormal.  actually it'll be tangent, binormal, normal for UVW, texcoords u,v and normal w
		float handedness;	//whether we're operating in 
	};

	std::vector<Vertex> vtx;
	std::vector<int>	index;

	void addVtx(float x,float y, float z, float tu, float tv) {
		Vertex v;
		v.pos.set(x,y,z);
		v.tc.set(tu,tv);
		//how will the basis be calculated?  using the u/v coordinates along with the normal
		//but how will the texcoords determine the tangent and binormal vectors?
		//after all, wont the basii be u/v basis be different for every single triangle touching this vertex?
		vtx.push_back(v);
	}

	void addTri(int i, int j, int k) {
		index.push_back(i);
		index.push_back(j);
		index.push_back(k);
	}

	
	void computeBasii() {
		// going to http://www.terathon.com/code/tangent.html for support

		int e;

		for (e = 0; e < (int)vtx.size(); e++) {
			vtx[e].basis = 0;
		}

		for (e = 0; e <	(int)index.size(); e+=3) {
			int i1 = index[e];
			int i2 = index[e+1];
			int i3 = index[e+2];
			Vertex &i = vtx[index[e]];
			Vertex &j = vtx[index[e+1]];
			Vertex &k = vtx[index[e+2]];

			vec3f &v1 = vtx[i1].pos;
			vec3f &v2 = vtx[i2].pos;
			vec3f &v3 = vtx[i3].pos;

			vec2f &w1 = vtx[i1].tc;
			vec2f &w2 = vtx[i2].tc;
			vec2f &w3 = vtx[i3].tc;

			vec3f dv1 = v2 - v1;
			vec3f dv2 = v3 - v1;

			vec2f dw1 = w2 - w1;
			vec2f dw2 = w3 - w1;

			//this looks like an inverse-matrix-multiplication to me...
			float r = 1.f / vec2det(dw1,dw2);
			//since we're no longer using dw, pre-scale it here:
			dw1 *= r;
			dw2 *= r;
			vec3f sdir = -dw1.y * dv2 + dw2.y * dv1;
			vec3f tdir =  dw1.x * dv2 - dw2.x * dv1;

			//add to the x axis of the basis
			i.basis.vec(0) += sdir;
			j.basis.vec(0) += sdir;
			k.basis.vec(0) += sdir;

			//add to the y axis of the basis
			i.basis.vec(1) += tdir;
			j.basis.vec(1) += tdir;
			k.basis.vec(1) += tdir;

			//and our normals:
			vec3f ndir = vec3unitNormal( i.pos, j.pos, k.pos );
			i.basis.vec(2) += ndir;
			j.basis.vec(2) += ndir;
			k.basis.vec(2) += ndir;
		}

		for (e = 0; e < (int)vtx.size(); e++) {
			Vertex &v = vtx[e];
			vec3f &t = v.basis.vec(0);	//s-axis / tangent
			vec3f &b = v.basis.vec(1);	//t-axis / binormal
			vec3f &n = v.basis.vec(2);	//normal
		    
			//normalize the normal
			n.normalize();

			// Gram-Schmidt orthogonalize
			(t -= (n * dot(t, n))).normalize();

			//since tangent 't' still lies in the same plane as <t,n>, we dont have to worry about
			//recalculating 'b' (the binormal, or basis.vec(1))
			//but just in case...
			b.normalize();
			//i'm voting on 2 or 0 -- either +1 or -1 the same vector
			//cout << "difference between accumulated and calculated binormal: " << vec3LengthLInf(b - (n * t)) << endl;
		    
			// Calculate handedness of the basis
			v.handedness = dot(cross(n, t), vtx[e].basis.vec(1)) < 0 ? -1 : 1;
			//ORRR just recalculate the binormal vector
			//it should be (due to the fact that before it already should align)
			//+- the same as the recalculated vector
			b = cross(n,t).normalize();
		}
	}

	void render() {
		glBegin(GL_TRIANGLES);
		for (int i = 0; i < (int)index.size(); i++) {
			Vertex &v = vtx[index[i]];

			glMultiTexCoord2fvARB(GL_TEXTURE0, v.tc);
			//so here we pass the normal through
			//how should we pass the tangent & binormal through?  other texcoords?  only one, and cross-product for the other?
			glNormal3fv(v.basis.vec(2));	//'v' is used interestingly in mat33f...
			//now feed in m.vec(0) through the unit1 coord
			glMultiTexCoord4fARB(GL_TEXTURE1, VEC3ELEM(v.basis.vec(0)), v.handedness);
			
			//and send it!
			glVertex3fv(v.pos);
		}
		glEnd();
	}
};

#ifdef ALLOW_DRAG_AND_DROP
LRESULT WINAPI DropFilesProc(int nCode, WPARAM wParam, LPARAM lParam);
HHOOK dropFilesHookHandle;
#endif //ALLOW_DRAG_AND_DROP

//communication between gui and shader scene
//its either this or we separate ALL the shader scene methods and headers...
class ShaderScene_Gui_Comm {
public:
	virtual void resetFragmentShaderUniforms() = 0;
	virtual void regenFragmentShader() = 0;
	virtual void refreshBumpMap() = 0;
	virtual void loadColorMap(const char *fn) = 0;
	virtual void loadBumpMap(const char *fn) = 0;
};

class ShaderGui;
static ShaderGui *mainGui = NULL;

class ShaderGui : public TextListener, public CheckboxListener {
protected:

#ifdef ALLOW_DRAG_AND_DROP
	HWND hwnd;
#endif //ALLOW_DRAG_AND_DROP

	//callback:
	ShaderScene_Gui_Comm *comm;

	enum {
		PREPROC_PARALLAX,
		PREPROC_OFFSET_LIMITING_EYE,
		PREPROC_ITERATIVE,
		PREPROC_ITERATIVE_LERP,
		PREPROC_SHADOW_TRACE,
		PREPROC_OFFSET_LIMITING_LIGHT,
		PREPROC_COUNT
	};

	static const char *preprocNames[PREPROC_COUNT];
	static const char *preprocMacros[PREPROC_COUNT];

	/*
	what kinds do we want to support:
	1) basic bumpmapping:
	2) basic parallax: PARALLAX
	3) basic parallax with offset limiting: PARALLAX, OFFSET_LIMIT
	4) iterative parallax: PARALLAX, ITERATIVE
	5) iterative parallax with offset limiting: PARALLAX, OFFSET_LIMIT, ITERATIVE
	6) iterative parallax with LERP: PARALLAX, ITERATIVE, ITERATIVE_LERP
	7) iterative parallax with offset limiting and LERP: PARALLAX, OFFSET_LIMIT, ITERATIVE, ITERATIVE_LERP

	*/

	stringstream preprocMacroBuffer;
	int preprocFlags;
	int parallaxSteps;	//app-specific. default 32

	float bumpHeight;	//call comm->refreshBumpMap when this is refreshed
	float polyThickness;	//comm->resetFragmentShaderUniforms when this is refreshed

	void buildPreprocMacroString() {
		//rebuild our preproc string
		preprocMacroBuffer.str("");

		for (int i = 0; i < PREPROC_COUNT; i++) {
			if (preprocFlags & bitflag(i)) {
				preprocMacroBuffer << "#define " << preprocMacros[i] << "\n";
			}
		}
		//finally, the step count:
		preprocMacroBuffer << "#define ITERATIVE_STEPS	" << parallaxSteps << ".0\n";

		//now tell the scene to regen our shader.
		comm->regenFragmentShader();
	}

	virtual void callback_onSetChecked(Checkbox* m, bool isChecked) {
		for (int i = 0; i < checkboxes.size(); i++) {
			if (m != checkboxes[i]) continue;
			preprocFlags &= ~bitflag(i);
			if (isChecked) preprocFlags |= bitflag(i);
			buildPreprocMacroString();
		}
	}

	//TextField callback:
	virtual void callback_onTextChange(TextField* m) {
		if (m == inputSteps) {
			int steps = atoi(m->getText().c_str());
			if (steps <= 0) return;
			parallaxSteps = steps;
			buildPreprocMacroString();	//which in turn regens the fragment shader
		} else if (m == inputPolyThick) {
			polyThickness = (float)atof(m->getText().c_str());
			cout << " polyThickness = " << polyThickness << endl;
			comm->resetFragmentShaderUniforms();
		} else if (m == inputBumpHeight) {
			bumpHeight = (float)atof(m->getText().c_str());
			cout << " bumpHeight = " << bumpHeight << endl;
			comm->refreshBumpMap();
		}
	}

	class ShaderTexDisplay : public pwn::Object {
		typedef pwn::Object super;
	public:
		static float getBoxSize() { return 4.f; }
		ShaderTexDisplay(const vec2f &_pos) : super(_pos, vec2f(getBoxSize())) { backgroundColor = 0xFFFFFFFF; }

		virtual bool display() {
			glPushAttrib(GL_ALL_ATTRIB_BITS);	//blendfunc?
			glDisable(GL_BLEND);
			bool ret = super::display();
			glPopAttrib();
			return ret;
		}
	};

	vector<gc::GlobalPointer<Checkbox> > checkboxes;	//[PREPROC_COUNT];
	gc::GlobalPointer<TextField> inputSteps, inputPolyThick, inputBumpHeight;
	gc::GlobalPointer<ShaderTexDisplay> colorTex, bumpTex;

	void reset() {
		//comm is untouched / persists
		preprocFlags = 0;
		parallaxSteps = 32;
		bumpHeight = 100.f;
		polyThickness = 0.5f;

		checkboxes.clear();

		inputSteps = NULL;
		inputPolyThick = NULL;
		inputBumpHeight = NULL;
		colorTex = NULL;
		bumpTex = NULL;
	}
public:

	ShaderGui(ShaderScene_Gui_Comm *_comm)
	:	comm(_comm),
		inputSteps(gc::sys->createPtr<TextField>()),
		inputPolyThick(gc::sys->createPtr<TextField>()),
		inputBumpHeight(gc::sys->createPtr<TextField>()),
		colorTex(gc::sys->createPtr<ShaderTexDisplay>()),
		bumpTex(gc::sys->createPtr<ShaderTexDisplay>())
	{
		reset();
	}
	~ShaderGui() { destroy(); }

	void destroy() {
		checkboxes.clear();
		inputPolyThick = NULL;
		inputBumpHeight = NULL;
		colorTex = NULL;
		bumpTex = NULL;
		
		menuSys.shutdown();

		//clear all 
		reset();
	}

	bool init() {
		mainGui = this;	//hack
		try {

#ifdef ALLOW_DRAG_AND_DROP
			//init drag and drop
			hwnd = winSys.getHWnd();
			if (!hwnd) {
				cout << "DragAndDrop.init(): couldn't find main window " << endl;
				return false;
			}
			DragAcceptFiles(hwnd, true);
			dropFilesHookHandle = SetWindowsHookEx(WH_GETMESSAGE, DropFilesProc, (HINSTANCE)NULL, GetCurrentThreadId());
			if (!dropFilesHookHandle) throw Exception() << "SetWindowsHookEx failed!";
#endif //ALLOW_DRAG_AND_DROP

			menuSys.init();
			menuSys.getRoot()->setScale(vec2f(12,12));


			//now that we're done with the menuSys, load up our Gui

			assert(!checkboxes.size());
			float y = 1.f;
			for (int i = 0; i < PREPROC_COUNT; i++) {
				checkboxes.push_back(gc::sys->createPtr(new Checkbox(vec2f(1,y), preprocNames[i])));
				if (!checkboxes[i]) throw Exception() << " couldnt create checkbox " << preprocNames[i];
				checkboxes[i]->addCheckboxListener(this);
				menuSys.getRoot()->addChild(checkboxes[i]);
				y += 1.1f;
			}
			y += 1.1f;

			menuSys.getRoot()->addChild(new Text(vec2f(1,y), "Iteration Steps:"));
			y += 1.1f;
			inputSteps = new TextField(vec2f(1,y), 10);
			if (!inputSteps) throw Exception() << " couldnt create input textfield";
			inputSteps->addTextListener(this);	//the java way
			{
				ostringstream s;
				s << parallaxSteps;
				inputSteps->setText(s.str());
			}
			menuSys.getRoot()->addChild(inputSteps);
			y += 2.1f;

			menuSys.getRoot()->addChild(new Text(vec2f(1,y), "Poly Depth:"));
			y += 1.1f;
			inputPolyThick = new TextField(vec2f(1,y), 10);
			if (!inputPolyThick) throw Exception() << " couldnt create input textfield";
			inputPolyThick->addTextListener(this);	//the java way
			{
				ostringstream s;
				s << polyThickness;
				inputPolyThick->setText(s.str());
			}
			menuSys.getRoot()->addChild(inputPolyThick);
			y += 2.1f;

			menuSys.getRoot()->addChild(new Text(vec2f(1,y), "Bumpmap Scale:"));
			y += 1.1f;
			inputBumpHeight = new TextField(vec2f(1,y), 10);
			if (!inputBumpHeight) throw Exception() << " couldnt create input textfield";
			inputBumpHeight->addTextListener(this);	//the java way
			{
				ostringstream s;
				s << bumpHeight;
				inputBumpHeight->setText(s.str());
			}
			menuSys.getRoot()->addChild(inputBumpHeight);
			y += 2.1f;

			colorTex = new ShaderTexDisplay(vec2f(2.f,y));
			if (!colorTex) throw Exception() << "couldn't create colorTex";
			menuSys.getRoot()->addChild(colorTex);
			bumpTex = new ShaderTexDisplay(vec2f(1.6f + 2.f * ShaderTexDisplay::getBoxSize(),y));
			if (!bumpTex) throw Exception() << "couldn't create bumpTex";
			menuSys.getRoot()->addChild(bumpTex);
			y += ShaderTexDisplay::getBoxSize() + 1.1f;

		} catch (Exception &e) {
			cout << "ShaderGui " << e << " failed!" << endl;
			return false;
		}

		return true;
	}

	void updateTextureIDs(int colorID, int bumpID) {
		assert(colorTex && bumpTex);
		colorTex->setTextureID(colorID);
		bumpTex->setTextureID(bumpID);
	}


	void display() {
		//overlay
		glDisable(GL_CULL_FACE);
		glDisable(GL_SCISSOR_TEST);
		menuSys.display();
	}

	bool getInput() { return menuSys.getInput(); }

	const string getPreprocMacroString() { return preprocMacroBuffer.str(); }

	float getBumpHeight() { return bumpHeight; }
	float getPolyThickness() { return polyThickness; }

#ifdef ALLOW_DRAG_AND_DROP
	void processDropMsg(MSG *msg) {
		cout << " got a drop! " << endl;
		DWORD lpos = GetMessagePos();
		POINT point;
		point.x = LOWORD(lpos);
		point.y = HIWORD(lpos);
		ScreenToClient(hwnd, &point);
		vec2f mpos((float)point.x / (float)disp.getSize().x,
					(float)point.y / (float)disp.getSize().y);
		cout << " mouse coords: " << mpos << endl;
		cout << " mouse obj coords: " << mouse.getFPos() << endl;
		mpos.x *= menuSys.getSize().x;
		mpos.y *= menuSys.getSize().y;
		cout << "scaled: " << mpos << endl;
		ShaderTexDisplay *buttons[2] = {colorTex, bumpTex};
		for (int i = 0; i < numberof(buttons); i++) {
			ShaderTexDisplay *button = buttons[i];
			box2f bbox;
			bbox.min = button->getAbsPos();
			bbox.max = bbox.min + button->getSize();
			cout << "box min: " << bbox.min << " max: " << bbox.max << endl;
			if (!bbox.contains(mpos)) continue;
			cout << "button " << i << " got the drop!" << endl;

			//extract the filename
			wchar_t filename[256];
			{
				HDROP hDrop = (HDROP)msg->wParam;
				int fileCount = DragQueryFile(hDrop, 0xFFFFFFFF, 0, 0);
				for (int i = 0; i < fileCount; i++) {
					memset(filename, 0, sizeof(filename));
					DragQueryFile(hDrop, i, filename, sizeof(filename));
					break;
				}
			}
			DragFinish((HDROP)msg->wParam);


			switch (i) {
				case 0:	//color map -- reload tex
					comm->loadColorMap(wstr2str(filename).c_str());
					break;
				case 1: //
					comm->loadBumpMap(wstr2str(filename).c_str());
					break;
			}
		}
	}
#endif //ALLOW_DRAG_AND_DROP
};

//I would put this next to the enum's def if C++ would let me . . .
const char *ShaderGui::preprocNames[ShaderGui::PREPROC_COUNT] = {
	"Parallax",
	"Offset Limit Eye",
	"Iterative",
	"Iterative LERP",
	"Shadow Trace",
	"Offset Limit Light"
};
const char *ShaderGui::preprocMacros[ShaderGui::PREPROC_COUNT] = {
	"PARALLAX",
	"OFFSET_LIMITING_EYE",
	"ITERATIVE",		//ITERATIVE_STEPS
	"ITERATIVE_LERP",
	"SHADOW_TRACE",
	"OFFSET_LIMITING_LIGHT"
};

#ifdef ALLOW_DRAG_AND_DROP
LRESULT WINAPI DropFilesProc(int nCode, WPARAM wParam, LPARAM lParam)  {
	MSG *msg = (MSG *)lParam;
	if (msg->message == WM_DROPFILES) {
		mainGui->processDropMsg(msg);
	}
	return CallNextHookEx(dropFilesHookHandle, nCode, wParam, lParam);
}
#endif //ALLOW_DRAG_AND_DROP

/*
here are my extensions to play with:
GL_ARB_multitexture
GL_EXT_texture_env_add
GL_EXT_compiled_vertex_array
GL_S3_s3tc
GL_ARB_depth_texture
GL_ARB_fragment_program
GL_ARB_fragment_program_shadow
GL_ARB_fragment_shader
GL_ARB_multisample
GL_ARB_occlusion_query
GL_ARB_point_parameters
GL_ARB_shader_objects
GL_ARB_shading_language_100
GL_ARB_shadow
GL_ARB_shadow_ambient
GL_ARB_texture_border_clamp
GL_ARB_texture_compression
GL_ARB_texture_cube_map
GL_ARB_texture_env_add
GL_ARB_texture_env_combine
GL_ARB_texture_env_crossbar
GL_ARB_texture_env_dot3
GL_ARB_texture_mirrored_repeat
GL_ARB_transpose_matrix
GL_ARB_vertex_blend
GL_ARB_vertex_buffer_object
GL_ARB_vertex_program
GL_ARB_vertex_shader
GL_ARB_window_pos
GL_ATI_draw_buffers
GL_ATI_element_array
GL_ATI_envmap_bumpmap
GL_ATI_fragment_shader
GL_ATI_map_object_buffer
GL_ATI_separate_stencil
GL_ATI_texture_env_combine3
GL_ATI_texture_float
GL_ATI_texture_mirror_once
GL_ATI_vertex_array_object
GL_ATI_vertex_attrib_array_object
GL_ATI_vertex_streams
GL_ATIX_texture_env_combine3
GL_ATIX_texture_env_route
GL_ATIX_vertex_shader_output_point_size
GL_EXT_abgr
GL_EXT_bgra
GL_EXT_blend_color
GL_EXT_blend_func_separate
GL_EXT_blend_minmax
GL_EXT_blend_subtract
GL_EXT_clip_volume_hint
GL_EXT_draw_range_elements
GL_EXT_fog_coord
GL_EXT_multi_draw_arrays
GL_EXT_packed_pixels
GL_EXT_point_parameters
GL_EXT_rescale_normal
GL_EXT_secondary_color
GL_EXT_separate_specular_color
GL_EXT_shadow_funcs
GL_EXT_stencil_wrap
GL_EXT_texgen_reflection
GL_EXT_texture3D
GL_EXT_texture_compression_s3tc
GL_EXT_texture_cube_map
GL_EXT_texture_edge_clamp
GL_EXT_texture_env_combine
GL_EXT_texture_env_dot3
GL_EXT_texture_filter_anisotropic
GL_EXT_texture_lod_bias
GL_EXT_texture_mirror_clamp
GL_EXT_texture_object
GL_EXT_texture_rectangle
GL_EXT_vertex_array
GL_EXT_vertex_shader
GL_HP_occlusion_test
GL_NV_texgen_reflection
GL_NV_blend_square
GL_NV_occlusion_query
GL_NV_point_sprite
GL_ARB_point_sprite
GL_SGI_color_matrix
GL_SGIS_texture_edge_clamp
GL_SGIS_texture_border_clamp
GL_SGIS_texture_lod
GL_SGIS_generate_mipmap
GL_SGIS_multitexture
GL_SUN_multi_draw_arrays
GL_WIN_swap_hint
WGL_EXT_extensions_string
WGL_EXT_swap_control
*/

class ShaderTestModule : public BaseModule, public ShaderScene_Gui_Comm {
	typedef BaseModule super;
protected:
	//static - one per
	VertexShader vertexShader;

	//one obj for our program, but we'll be regen'ing it lots
	FragmentShader fragmentShader;

	//here's our fragment shader string
	string fragmentShaderCode;

	//dynamic - lots per
	ShaderProgram shaderProgram;

	//shader gui
	ShaderGui *gui;

	//light source:
	vec4f lightpos;

	//zero any pointers / default any values *HERE*
	void reset() {
		gui = NULL;
		lightpos.set(-7.60227, 0.788751, 0.296831, 1);
	}

public:

	Texture tex, bumpsrc, bump;
	Model3D mdl;

	virtual void resetFragmentShaderUniforms() {
		shaderProgram.setUniform1i("colorMap", 0);		//colormap <=> unit 0
		shaderProgram.setUniform1i("normalMap", 1);		//bumpmap <=> unit 1
		shaderProgram.setUniform1f("polyThickness", gui->getPolyThickness());	//polygon thickness (proportional to texture size or texel size?)
		ShaderProgram::useNone();
	}

	virtual void regenFragmentShader() {
		string buffer = gui->getPreprocMacroString();
		cout << "rebuilding with macro strings:" << endl << buffer << endl;
		buffer += fragmentShaderCode;
		fragmentShader.setSource(buffer.c_str());
		fragmentShader.compile();

		shaderProgram.link();
		shaderProgram.printLog();
		resetFragmentShaderUniforms();
	}

	virtual void loadColorMap(const char *fn) {
		if (!tex.load2D(fn)) {
			cout << "failed to load the texture map " << fn << endl;
		}
	}
	virtual void loadBumpMap(const char *fn) {
		if (!bumpsrc.load2D(fn)) {
			cout << "failed to load bumpmap " << fn << endl;
			return;
		}
		refreshBumpMap();
	}
	virtual void refreshBumpMap() {
		BumpMap::gen(bumpsrc, bump, gui->getBumpHeight());	//get the tangent-space texture.  why such high exhaggeration needed for lighting to look right?
	}


	//all my messy shader init code:
protected:

	void printInfoLog(GLhandleARB obj) {
		GLint infologLength = 0;
		GLint charsWritten  = 0;
		char *infoLog;

		glGetObjectParameterivARB(obj, GL_OBJECT_INFO_LOG_LENGTH_ARB, &infologLength);

		if (infologLength > 0) {
			infoLog = (char *)malloc(infologLength);
			glGetInfoLogARB(obj, infologLength, &charsWritten, infoLog);
			printf("%s\n",infoLog);
			free(infoLog);
		}
	}

public:

	ShaderTestModule() {
		reset();
	}

	virtual void init() {
		sys.setInitSize(vec2i(800, 600));
		super::init();
		con.font.load2D("font.bmp");

		glClearColor(0,0,0,0);

		//enable depth testing by default
		glEnable(GL_DEPTH_TEST);

	//	glEnable(GL_NORMALIZE);

		// setup our light model

		glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);	//slower, but more accurate
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, vec4f(0,0,0,1));

		// setup our light0

		glLightfv(GL_LIGHT0, GL_AMBIENT, vec4f(1,1,1,1));
		glLightfv(GL_LIGHT0, GL_DIFFUSE, vec4f(1,1,1,1));
		glLightfv(GL_LIGHT0, GL_SPECULAR, vec4f(1,1,1,1));
		glLightfv(GL_LIGHT0, GL_POSITION, vec4f(0,0,1,0));	//notice - this position is loaded relative to the identity modelview space, not our viewing space

		//glEnable(GL_LIGHT0);

		if (found_GL_EXT_separate_specular_color) {
			// enable separate specular color calculation and addition
			glLightModelf(GL_LIGHT_MODEL_COLOR_CONTROL_EXT, GL_SEPARATE_SPECULAR_COLOR_EXT);
		}

		// setup our color material - map glColor calls to the diffuse material parameter

	//	glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
	//	glEnable(GL_COLOR_MATERIAL);

		gui = new ShaderGui(this);
		if (!gui) throw Exception() << "ShaderGui";
		if (!gui->init()) throw Exception() << "ShaderGui init";

		//post-load-scene, reset our 3d view to sit behind the character
		const float viewscale =  0.1f;
		const float viewnear = _VIEW_ZNEAR * viewscale;
		const float viewfar = _VIEW_ZFAR * viewscale;
		base.views[VIEW_3D].set(
			this,
			false, 
			viewnear,
			viewnear,
			VIEW_ADJUST_HEIGHT,	//adjust the height based upon the width for sake of maintaining aspect ratio
			viewnear,
			viewfar,
			//vec3f(0,0,10),
			//quatf(0,0.5f*M_SQRT_2,0,0.5f*M_SQRT_2) * quatf(-0.5f*M_SQRT_2,0,0,0.5f*M_SQRT_2));	//looking down:
			//90 degrees on Y [then, relative to the view] -90 degrees on X
			vec3f(-55.306, 0.000232944, 14.3633),
			quatf(-0.700785, 0.094343, 0.0943405, 0.700785));

		base.camSpeed = 10;
		base.camRot = 7.5;

		//specify our model
		//should texcoords be associated with vertices or indices?
		//vertices.  we'll just have overlapping vertices wherever an index references the same pos with dif texcoords

		//TODO - provide your own model! and a loader too...
		{
			const int detail = 32;
			int e = 0;
			for (int j = 0; j < detail; j++) {
				for (int i = 0; i < detail; i++, e++) {

					float x = (float)i / (float)(detail-1);
					float y = (float)j / (float)(detail-1);

					//float u = x*2.f - 1.f;
					//float v = y*2.f - 1.f;
					//float w = 1.f - u*u - v*v;
					//w = w < 0 ? 0 : sqrt(w);
					//float w = 0;
					//mdl.addVtx(4.f*u, 4.f*v, 4.f*w, 4.f*x, 4.f*y);

					//vec3f s(x, y, 0.f);
					vec3f s(1,0,0);
					//s = quatRotate( angleAxisToQuat( vec3f(0,1,0), y * 180.f ), s);
					//s = quatRotate( angleAxisToQuat( vec3f(0,0,1), x * 360.f ), s);
					s = mat33Rot<float>( vec3f(0,1,0), (2.f*y-1.f)*90.f) * s;
					s = mat33Rot<float>( vec3f(0,0,1), x*360.f) * s;
					s *= 4.f;
					vec2f t(x,y);
					t.x *= 32.f;
					t.y *= -16.f;
					mdl.addVtx(VEC3ELEM(s), VEC2ELEM(t));

					if (i < detail-1 && j < detail-1) {
						mdl.addTri(e, e + detail, e + detail + 1);
						mdl.addTri(e + detail + 1, e + 1, e);
					}
				}
			}
		}

		mdl.computeBasii();

#if 1
		loadColorMap("cheap brick.bmp");
		loadBumpMap("bumptest.bmp");
#else
		const char *texFileName = "cheap brick.bmp";
		if (!tex.load2D(texFileName)) {			//load the texture
			throw Exception() << "Texture::load2D(" << texFileName << ")";
		}

#if 0	//same bumpmap texture
		BumpMap::gen(tex, bump, 10.f);			//generate its associated bumpmap:
#else	//separate bumpmap texture
		{
			Texture bumpsrc;
			bumpsrc.load2D("bumptest.bmp");
			BumpMap::gen(bumpsrc, bump, 100.f);	//get the tangent-space texture.  why such high exhaggeration needed for lighting to look right?
		}	//bumpsrc dtor
#endif
#endif
		gui->updateTextureIDs(
			tex.getTextureID(),
			bump.getTextureID());

		//our object material:
		glMaterialfv(GL_FRONT, GL_EMISSION, vec4f(0,0,0,0));
		glMaterialfv(GL_FRONT, GL_AMBIENT, vec4f(0,0,0,1));
		glMaterialfv(GL_FRONT, GL_DIFFUSE, vec4f(1,1,1,1));
		glMaterialfv(GL_FRONT, GL_SPECULAR, vec4f(1,1,1,1));
		glMaterialf(GL_FRONT, GL_SHININESS, 10.f);

		Texture::unbind();

		//constructs shaderProgram.
		if (!shaderSystem.init()) throw Exception() << "shaderSystem.init()";

//#define SHADERFILENAME		"bumpmap"

		//offset parallax with a limiting option
//#define SHADERFILENAME		"parallax offset limit"

		//in the works
		//currently supports the following preprocessor options:
		//PARALLAX: for enabling parallax mapping
		//OFFSET_LIMIT: for enabling offset-limiting. depends on PARALLAX
		//ITERATIVE: for enabling iterative parallax mapping.  depends on PARALLAX
		//ITERATIVE_LERP: for LERPing back to correct artifacting.  depends on ITERATIVE
#define SHADERFILENAME		"iterative parallax"

		{
			const char *fragShaderFilename = "shaders/"SHADERFILENAME".frag";
			FileBuffer tf(fragShaderFilename);
			if (!tf.getData()) throw Exception() << "Frag shader filename " << fragShaderFilename << " has no data!";
			if (!tf.getSize()) throw Exception() << "Frag shader filename " << fragShaderFilename << " is empty!";
			fragmentShaderCode = tf.getData();	//accum once.
		}	//FileBuffer destructor

		vertexShader.createFromSource("shaders/basic bumpmap.vert");
		fragmentShader.create();
		fragmentShader.setSource(fragmentShaderCode.c_str());
		fragmentShader.compile();

		shaderProgram.create();
		shaderProgram.attach(fragmentShader);
		shaderProgram.attach(vertexShader);

		regenFragmentShader();
	}
	
	virtual void shutdown() {
		if (gui) delete gui;
		
		//call these while a gl context exists
		//or else the static dtor will try to unload them when it doesn't
		//and the program will crash on exit
		//TODO - check in the dtor for that context?
		tex.unload();
		bumpsrc.unload();
		bump.unload();
		vertexShader.destroy();
		fragmentShader.destroy();
		shaderProgram.destroy();

		reset();

		super::shutdown();
	}

	//todo - make a window callback for display and put this in its code ... ?
	virtual void update() {

		//if we didnt get any gui interaction...
		if (!gui->getInput()) {
		
			if (base.mouseOverViewport) {
				Viewport *vp = base.mouseOverViewport;
				View *v = vp->getView();
				if (v) {

					//scene->getInput() ... moving the light around and what not ... ?
					if (mouse.getButtonState(mouse.BUTTON_LEFT) == BUTTON_STATE_DOWN && (mouse.getMove().x || mouse.getMove().y)) {

						//if there are no mouse clicks then this becomes extra work every frame . . .
						//here's a std mouse transform code:
						vec3f mouseThis = vp->getPixelPos(mouse.getPos());
						vec3f mouseLast = vp->getPixelPos(mouse.getPos() - mouse.getMove());
						vec3f mouseDelta = mouseThis - mouseLast;

						//move the light position
						vec3f origin = v->getPos();
						vec3f &dest = *(vec3f*)&lightpos;

						//now scale these vectors a distance of 'goal - origin' dot forward
						float scaleDist = dot(dest - origin, quatXAxis(v->getAngle()));	//assuming v's angle is unit-ized
						mouseDelta *= scaleDist / v->getZNear();
						

						//replace key'z' with mouse scroll wheel delta Z ... once we find how to read it while in GLUT . . .
						//has the keyboard handler incorperated shift states?
						//those are typically a pain to fully extract from GLUT...
						if (kbd.keystate[Key::LSHIFT]) {
							//take the up-component, turn it into the forward/backwards component
							mouseDelta = quatXAxis(v->getAngle()) * dot(mouseDelta, quatZAxis(v->getAngle()));
						}

						dest += mouseDelta;
					}
				}
			}
		}

		//typical update stuff:
		//todo: run a lua shell on the console, or as an overlay console, based on the new menu system
		//todo: redesign the mouselib over the mathlib
	}
	virtual void render(const Viewport *vp, const View *v) {

#if 0
		mat33f m;
		glBegin(GL_LINES);
		for (int i = 0; i < 3; i++) {
			glColor3fv(m.vec(i).v);
			glVertex3f(0,0,0);
			glVertex3fv(m.vec(i).v);
		}
		glEnd();
#endif

		//update in space . . .
		//the w=1 is important since opengl will multiply this by modelview as soon as you set it
		glLightfv(GL_LIGHT0, GL_POSITION, lightpos);

		glColor3f(1,1,1);
		glPointSize(5.f);
		glBegin(GL_POINTS);
		glVertex3fv(lightpos);
		glEnd();
		glPointSize(1.f);

		glEnable(GL_CULL_FACE);

		tex.bind(0);
		bump.bind(1);
		shaderProgram.use();

		mdl.render();

		ShaderProgram::useNone();

		//reset the active texture unit while we're at it
		Texture::unbindAll();

		//menu system
		gui->display();
	}
};
static ShaderTestModule shaderTestModule;
BaseModule *RegisterApplication() { return &shaderTestModule; }
